/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under both the MIT license found in the
 * LICENSE-MIT file in the root directory of this source tree and the Apache
 * License, Version 2.0 found in the LICENSE-APACHE file in the root directory
 * of this source tree.
 */

/// This module provides a check for auto-imported functions from the `erlang` module
/// module.
///
pub fn in_erlang_module(f: &str, a: usize) -> bool {
    is_erlang_fun(f, a) || is_erlang_type(f, a)
}

pub fn is_erlang_fun(f: &str, a: usize) -> bool {
    match (f, a) {
        // The erlang compiler sys_core_fold.erl file has a function
        // `is_auto_imported/3` which calls `erl_internal:bif/2`.
        // This list is derived from `erl_internal:bif/2`.
        ("abs", 1) => true,
        ("alias", 0) => true,
        ("alias", 1) => true,
        ("apply", 2) => true,
        ("apply", 3) => true,
        ("atom_to_binary", 1) => true,
        ("atom_to_binary", 2) => true,
        ("atom_to_list", 1) => true,
        ("binary_part", 2) => true,
        ("binary_part", 3) => true,
        ("binary_to_atom", 1) => true,
        ("binary_to_atom", 2) => true,
        ("binary_to_existing_atom", 1) => true,
        ("binary_to_existing_atom", 2) => true,
        ("binary_to_integer", 1) => true,
        ("binary_to_integer", 2) => true,
        ("binary_to_float", 1) => true,
        ("binary_to_list", 1) => true,
        ("binary_to_list", 3) => true,
        ("binary_to_term", 1) => true,
        ("binary_to_term", 2) => true,
        ("bitsize", 1) => true,
        ("bit_size", 1) => true,
        ("bitstring_to_list", 1) => true,
        ("byte_size", 1) => true,
        ("ceil", 1) => true,
        ("check_old_code", 1) => true,
        ("check_process_code", 2) => true,
        ("check_process_code", 3) => true,
        ("date", 0) => true,
        ("delete_module", 1) => true,
        ("demonitor", 1) => true,
        ("demonitor", 2) => true,
        ("disconnect_node", 1) => true,
        ("element", 2) => true,
        ("erase", 0) => true,
        ("erase", 1) => true,
        ("error", 1) => true,
        ("error", 2) => true,
        ("error", 3) => true,
        ("exit", 1) => true,
        ("exit", 2) => true,
        ("float", 1) => true,
        ("float_to_list", 1) => true,
        ("float_to_list", 2) => true,
        ("float_to_binary", 1) => true,
        ("float_to_binary", 2) => true,
        ("floor", 1) => true,
        ("garbage_collect", 0) => true,
        ("garbage_collect", 1) => true,
        ("garbage_collect", 2) => true,
        ("get", 0) => true,
        ("get", 1) => true,
        ("get_keys", 0) => true,
        ("get_keys", 1) => true,
        ("group_leader", 0) => true,
        ("group_leader", 2) => true,
        ("halt", 0) => true,
        ("halt", 1) => true,
        ("halt", 2) => true,
        ("hd", 1) => true,
        ("integer_to_binary", 1) => true,
        ("integer_to_binary", 2) => true,
        ("integer_to_list", 1) => true,
        ("integer_to_list", 2) => true,
        ("iolist_size", 1) => true,
        ("iolist_to_binary", 1) => true,
        ("is_alive", 0) => true,
        ("is_process_alive", 1) => true,
        ("is_atom", 1) => true,
        ("is_boolean", 1) => true,
        ("is_binary", 1) => true,
        ("is_bitstring", 1) => true,
        ("is_float", 1) => true,
        ("is_function", 1) => true,
        ("is_function", 2) => true,
        ("is_integer", 1) => true,
        ("is_list", 1) => true,
        ("is_map", 1) => true,
        ("is_map_key", 2) => true,
        ("is_number", 1) => true,
        ("is_pid", 1) => true,
        ("is_port", 1) => true,
        ("is_reference", 1) => true,
        ("is_tuple", 1) => true,
        ("is_record", 2) => true,
        ("is_record", 3) => true,
        ("length", 1) => true,
        ("link", 1) => true,
        ("list_to_atom", 1) => true,
        ("list_to_binary", 1) => true,
        ("list_to_bitstring", 1) => true,
        ("list_to_existing_atom", 1) => true,
        ("list_to_float", 1) => true,
        ("list_to_integer", 1) => true,
        ("list_to_integer", 2) => true,
        ("list_to_pid", 1) => true,
        ("list_to_port", 1) => true,
        ("list_to_ref", 1) => true,
        ("list_to_tuple", 1) => true,
        ("load_module", 2) => true,
        ("make_ref", 0) => true,
        ("map_size", 1) => true,
        ("map_get", 2) => true,
        ("max", 2) => true,
        ("min", 2) => true,
        ("module_loaded", 1) => true,
        ("monitor", 2) => true,
        ("monitor", 3) => true,
        ("monitor_node", 2) => true,
        ("node", 0) => true,
        ("node", 1) => true,
        ("nodes", 0) => true,
        ("nodes", 1) => true,
        ("now", 0) => true,
        ("open_port", 2) => true,
        ("pid_to_list", 1) => true,
        ("port_to_list", 1) => true,
        ("port_close", 1) => true,
        ("port_command", 2) => true,
        ("port_command", 3) => true,
        ("port_connect", 2) => true,
        ("port_control", 3) => true,
        ("pre_loaded", 0) => true,
        ("process_flag", 2) => true,
        ("process_flag", 3) => true,
        ("process_info", 1) => true,
        ("process_info", 2) => true,
        ("processes", 0) => true,
        ("purge_module", 1) => true,
        ("put", 2) => true,
        ("ref_to_list", 1) => true,
        ("register", 2) => true,
        ("registered", 0) => true,
        ("round", 1) => true,
        ("self", 0) => true,
        ("setelement", 3) => true,
        ("size", 1) => true,
        ("spawn", 1) => true,
        ("spawn", 2) => true,
        ("spawn", 3) => true,
        ("spawn", 4) => true,
        ("spawn_link", 1) => true,
        ("spawn_link", 2) => true,
        ("spawn_link", 3) => true,
        ("spawn_link", 4) => true,
        ("spawn_request", 1) => true,
        ("spawn_request", 2) => true,
        ("spawn_request", 3) => true,
        ("spawn_request", 4) => true,
        ("spawn_request", 5) => true,
        ("spawn_request_abandon", 1) => true,
        ("spawn_monitor", 1) => true,
        ("spawn_monitor", 2) => true,
        ("spawn_monitor", 3) => true,
        ("spawn_monitor", 4) => true,
        ("spawn_opt", 2) => true,
        ("spawn_opt", 3) => true,
        ("spawn_opt", 4) => true,
        ("spawn_opt", 5) => true,
        ("split_binary", 2) => true,
        ("statistics", 1) => true,
        ("term_to_binary", 1) => true,
        ("term_to_binary", 2) => true,
        ("term_to_iovec", 1) => true,
        ("term_to_iovec", 2) => true,
        ("throw", 1) => true,
        ("time", 0) => true,
        ("tl", 1) => true,
        ("trunc", 1) => true,
        ("tuple_size", 1) => true,
        ("tuple_to_list", 1) => true,
        ("unalias", 1) => true,
        ("unlink", 1) => true,
        ("unregister", 1) => true,
        ("whereis", 1) => true,
        _ => false,
    }
}

pub fn is_erlang_type(f: &str, a: usize) -> bool {
    match (f, a) {
        // This part is from https://www.erlang.org/doc/reference_manual/typespec.html
        ("any", 0) => true,
        ("atom", 0) => true,
        ("float", 0) => true,
        ("fun", 1) => true,
        ("fun", 0) => true,
        ("integer", 0) => true,
        ("list", 1) => true,
        ("maybe_improper_list", 2) => true,
        ("none", 0) => true,
        ("nonempty_improper_list", 2) => true,
        ("nonempty_list", 1) => true,
        ("pid", 0) => true,
        ("port", 0) => true,
        ("reference", 0) => true,
        ("tuple", 0) => true,

        // This part is from table 7.1 of https://www.erlang.org/doc/reference_manual/typespec.html
        ("term", 0) => true,
        ("binary", 0) => true,
        ("nonempty_binary", 0) => true,
        ("bitstring", 0) => true,
        ("nonempty_bitstring", 0) => true,
        ("boolean", 0) => true,
        ("byte", 0) => true,
        ("char", 0) => true,
        ("nil", 0) => true,
        ("number", 0) => true,
        ("list", 0) => true,
        ("maybe_improper_list", 0) => true,
        ("nonempty_list", 0) => true,
        ("string", 0) => true,
        ("nonempty_string", 0) => true,
        ("iodata", 0) => true,
        ("iolist", 0) => true,
        ("map", 0) => true,
        ("function", 0) => true,
        ("module", 0) => true,
        ("mfa", 0) => true,
        ("arity", 0) => true,
        ("identifier", 0) => true,
        ("node", 0) => true,
        ("timeout", 0) => true,
        ("no_return", 0) => true,
        _ => false,
    }
}
